from typing import Generic, TypeVar

from rainbowadn.atomic import *
from rainbowadn.core import *
from .actions import *
from .core import *

__all__ = ('AVL',)

ActiveKeyType = TypeVar('ActiveKeyType')
TreeType = TypeVar('TreeType')


class AVL(BinaryBalancing[ActiveKeyType, Integer, TreeType]):
    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    async def metadata(
            self,
            treel: TreeType,
            treer: TreeType,
            key: HashPoint[ActiveKeyType],
            creation: BinaryCreation[ActiveKeyType, Integer, TreeType]
    ) -> HashPoint[Integer]:
        assert isinstance(creation, BinaryCreation)
        height_l, height_r = await gather(
            self.height(BinaryProtocolized(creation, treel)),
            self.height(BinaryProtocolized(creation, treer)),
        )
        return HashPoint.of(
            Integer(
                1
                +
                max(
                    height_l,
                    height_r,
                )
            )
        )

    @classmethod
    async def height(
            cls,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> int:
        return await HeightAction().on(protocolized)

    async def balance(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> TreeType:
        return await BalanceAction().on(protocolized)


class HeightAction(
    BinaryAction[ActiveKeyType, Integer, TreeType, int],
    Generic[ActiveKeyType, TreeType]
):
    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> int:
        return 0

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, Integer, TreeType]
    ) -> int:
        metadata: Integer = await case.split.metadata.resolve()
        return metadata.integer


class BalanceAction(
    BinaryAction[ActiveKeyType, Integer, TreeType, TreeType],
    Generic[ActiveKeyType, TreeType]
):
    async def on_null(
            self,
            protocolized: BinaryProtocolized[ActiveKeyType, Integer, TreeType],
    ) -> TreeType:
        return protocolized.tree

    async def on_split(
            self,
            case: ProtocolizedBinarySplit[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        height_l, height_r = await gather(
            AVL.height(case.protocolizedl()),
            AVL.height(case.protocolizedr()),
        )
        delta = height_l - height_r
        assert isinstance(delta, int)
        if delta < -1:
            return await self.on_symmetric(InnerOuter(case.protocol), case.split)
        elif delta > 1:
            return await self.on_symmetric(OuterInner(case.protocol), case.split)
        else:
            return case.tree

    @classmethod
    async def on_symmetric(
            cls,
            symmetry: Symmetric[ActiveKeyType, Integer, TreeType],
            split: BinarySplit[ActiveKeyType, Integer, TreeType]
    ) -> TreeType:
        assert isinstance(symmetry, Symmetric)
        assert isinstance(split, BinarySplit)
        splito = await symmetry.protocol.fsplit(symmetry.outer(split))
        height_oi, height_oo = await gather(
            AVL.height(symmetry.protocolizedi(splito)),
            AVL.height(symmetry.protocolizedo(splito)),
        )
        if height_oi > height_oo:
            splitoi = await symmetry.protocol.fsplit(symmetry.inner(splito))
            inner, outer = await gather(
                symmetry.tree(symmetry.inner(split), symmetry.inner(splitoi), split.key),
                symmetry.tree(symmetry.outer(splitoi), symmetry.outer(splito), splito.key),
            )
            return await symmetry.tree(
                inner,
                outer,
                splitoi.key
            )
        else:
            return await symmetry.tree(
                await symmetry.tree(symmetry.inner(split), symmetry.inner(splito), split.key),
                symmetry.outer(splito),
                splito.key
            )
