import functools
import time
from typing import Generic, TypeVar

__all__ = ('Instrumentation', 'Counter', 'Concurrency',)


IType = TypeVar('IType')


class Instrumentation(Generic[IType]):
    deinstrumentation = {}

    def __init__(self, target, methodname: str):
        self.target = target
        self.methodname = methodname

    def instrument(self, method, *args, **kwargs):
        raise NotImplementedError

    def __enter__(self: IType) -> IType:
        assert not hasattr(self, 'method')
        self.method = getattr(self.target, self.methodname)

        @functools.wraps(self.method)
        def wrap(*args, **kwargs):
            return self.instrument(self.method, *args, **kwargs)

        self.wrap = wrap

        setattr(self.target, self.methodname, self.wrap)

        return self

    def schedule_deinstrumentation(self):
        self.deinstrumentation[self.wrap] = self.method
        del self.wrap
        del self.method

    def deinstrument(self):
        while (method := getattr(self.target, self.methodname)) in self.deinstrumentation:
            setattr(self.target, self.methodname, self.deinstrumentation.pop(method))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.schedule_deinstrumentation()
        self.deinstrument()


class Counter(Instrumentation):
    def __init__(self, target, methodname: str):
        super().__init__(target, methodname)
        self.counter = 0

    def instrument(self, method, *args, **kwargs):
        self.counter += 1
        return method(*args, **kwargs)


class Concurrency(Instrumentation):
    start = time.time()

    def __init__(self, target, methodname: str):
        super().__init__(target, methodname)
        self.concurrency = 0
        self.logs: list[tuple[float, int]] = []

    def time(self) -> float:
        return time.time() - self.start

    async def instrument(self, method, *args, **kwargs):
        self.logs.append((self.time(), self.concurrency))
        self.concurrency += 1
        self.logs.append((self.time(), self.concurrency))
        result = await method(*args, **kwargs)
        self.logs.append((self.time(), self.concurrency))
        self.concurrency -= 1
        self.logs.append((self.time(), self.concurrency))
        return result
